/*
 * TelPac-Node/LinuX developed by Hans-J. Barthen, DL5DI based on tcp_call
 * from ax25-tools 0.0.8, enhanced for full-binary transfer and automatic
 * password protected user- and node-logon at Winlink 2000 servers.
 *
 * This code is like the original tools term of the GNU licence.
 *
 * In case of problems please contact dl5di@darc.de/dl5di@db0lj.ampr.org
 * Hans-Juergen Barthen, Franz-Josef-Str. 20, D-56642 Kruft
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <time.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <fcntl.h>
#include "../user_io.h"


void alarm_handler(int sig)
/*
This did not work because a timeout does not stop the connect statement
*/
{
 signal(SIGALRM, alarm_handler); // reset signal handler
}

int node_connect(char *hhost, char *hport)

/***************************************************************************
*
* Suggested modication by Andy, K3JSE
*
* Try connecting to host, on success return socket, on failure return -1 
*
* Set socket to non blocking and call connect
* select waits for either
*  - timeout which signals probable connect error
*  - ready to write which signals connection successful
* set socket socket back to non blocking and return
*
***************************************************************************/
  {
  struct timeval tv;
  fd_set writefds;
  long arg;
  int res, mysock;
  struct sockaddr_in node_addr;
  struct hostent *he;  
  char mybuf[80];

  if ((mysock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
    //perror("Create Socket");
    err("Could not create socket"); // this is a fatal error!
    }

  sprintf(mybuf,"INFO: Host Name %s, Port %s \r\n", hhost, hport);
  send(STDOUT_FILENO, mybuf, strlen(mybuf), 0);          

  if ((he=gethostbyname(hhost)) == NULL)
   {  // get the host info
   sprintf(mybuf,"ERROR: Unknown host %s. Try next line.\r\n", hhost);
   send(STDOUT_FILENO, mybuf, strlen(mybuf), 0);             
   return(-1);
   }

  node_addr.sin_family = AF_INET;
  node_addr.sin_port = htons(atoi(hport));
  node_addr.sin_addr = *((struct in_addr *)he->h_addr);
  bzero(&(node_addr.sin_zero), 8);

  // Set socket to non-blocking, preserve other flags 
  arg = fcntl(mysock, F_GETFL, NULL); 
  arg |= O_NONBLOCK; 
  fcntl(mysock, F_SETFL, arg);   

  res = connect(mysock, (struct sockaddr *)&node_addr, sizeof(struct sockaddr));

  if (res < 0)
    {
    if (errno == EINPROGRESS)
      { // expected result, waiting for write ready or timeout
      while(1)
        {
        tv.tv_sec = 5; // timeout set to 5 seconds for testing.........
        tv.tv_usec = 0; 
        FD_ZERO(&writefds);       
        FD_SET(mysock,&writefds);        
        res = select(mysock+1, NULL, &writefds, NULL, &tv);       
        if (res > 0) // selected for write
          {
          sprintf(mybuf,"Success: Connected\r\n");
          send(STDOUT_FILENO, mybuf, strlen(mybuf), 0);          
          // Set back to blocking mode... 
          arg = fcntl(mysock, F_GETFL, NULL); 
          arg &= ~O_NONBLOCK; 
          fcntl(mysock, F_SETFL, arg);           
          return(mysock);
          }
        else
          {
          //printf("Ouch - Timed out\n");
          sprintf(mybuf,"ERROR: Timeout connecting to node %s. Try next line\r\n", hhost);
          send(STDOUT_FILENO, mybuf, strlen(mybuf), 0);                    
          close(mysock);
          return(-1);          
          }  
        } // end while (1)
      }  // end erro == EINPROGRESS
    else // something bad happened
      {
      // probably should kill the process here!
      sprintf(mybuf,"FATAL ERROR: Selecting\r\n");
      send(STDOUT_FILENO, mybuf, strlen(mybuf), 0);                          
      close(mysock);
      return(-1);       
      }
    } // end res < 0  
  // connected immediately....... 
  return(mysock);
  
  }

int main(int argc, char **argv)
{
	char revision[80];
	char buffer[256];
	char inputline[256];
	char nodecall[20];
	char passwd[30];
	char hostname[30];
	int s;  //, addrlen = sizeof(struct sockaddr_in);
	int verbose = 1;
//	struct sockaddr_in addr;
//	struct hostent *hp;
//	struct servent *sp;
	struct ust userstat;
	time_t tstart;
	time_t tstop;
	int status = 0;
	int i;
	char *usercall = '\0';
	char *cp;
	char *hhost;
	char *hport;
	char *hpass;
	FILE *config_file;
	FILE *host_file;

	paclen = 128;  // probably should default to a "safe value" like 128
	
	sprintf(revision,"\rTelPac-Node/LinuX %s (dl5di %s)\r\r",REV,__DATE__);

	while ((s = getopt(argc, argv, "o:p:q")) != -1) {
		switch (s) {
		case 'o':
		case 'p':
			paclen = atoi(optarg);
			break;
		case 'q':
			verbose = 0;
			break;
		case ':':
		case '?':
			err("ERROR: invalid option usage\r\n");
			return 1;
		}
	}

	if (paclen < 1) {
		err("ERROR: invalid paclen\r\n");
	}

	/*
	 * Argument should be "telpac_node usercall"
	 */
	if ((argc - optind) != 1) {
		err("ERROR: invalid number of parameters\r\n");
	}


	/*
	 * Prepare the user call.
	 */
	usercall = (argv[optind]);
	if (strlen(usercall) > 9) {
	    err("ERROR ! Invalid User-Callsign !\r\n");
	} else {
	    for (i=0; i < strlen(usercall); i++) {
		usercall[i] = toupper(usercall[i]);
		if ((usercall[i] == '-') && (usercall[i+1] == '0')){
		    usercall[i] = '\0';
		    break;
		}
	    }
	}

/*
 * read hostfile (telpac.hst), format "hostname:port:password"
 * try to connect host 
 * loop to other entries if connection fails
 */

	if((host_file = fopen(hostfile,"r")) != NULL) {

	    while(fgets(inputline, 80, host_file) != NULL) {
		cp = inputline;

		/* Jump over comments */
		if((*cp == '#') || (*cp == ';')) {
		    continue;
		}
		
		hhost = inputline;
		hport = NULL;
		hpass = NULL;

		/* search for start of 2nd token / port */
		while((*cp != '\0') && (*cp != '\r') && (*cp != '\n')) {
		    if(*cp == ':') {
			*cp = '\0';
			cp++;
			hport = cp;
			break;
		    }
		    cp++;
		}
		if (strlen(hhost) == 0 ) {
		    err("ERROR: missing hostname in configfile\r\n");
		}


		/* search for start of 3rd token / password */
		while((*cp != '\0') && (*cp != '\r') && (*cp != '\n')){
		    if(*cp == ':') {
			*cp = '\0';
			cp++;
			hpass = cp;
			break;
		    }
		    cp++;
		}

		if (strlen(hport) == 0 ) {
		    sprintf(buffer,"ERROR: missing service '%s'. Try next line.\n", hport);
		    msg(buffer);
		    status++;
		    continue;
		}


		/*
		 * Now the rest must be a password.
		 */

		while(*cp != '\0'){
		    if((*cp == '\r') || (*cp == '\n')){
			*cp = '\0';
			break;
		    }
		    cp++;
		}

		if (strlen(hpass) == 0 ) {
		    sprintf(buffer,"ERROR: missing password at host '%s'. Try next host.\n",hhost);
		    msg(buffer);
		    status++;
		    continue;
		}

		strcpy(passwd,hpass);
		strcpy(hostname,hhost);

                // Try connecting to node
                
                s = node_connect(hhost, hport);
                
                if (s == -1)  // connection failed
                  status++;   // increment connection counter

		if(s > 0) {
		// We are connected
                    status = 0; // must do this or line below exits  		
		    break;
		}

	    }	/* end of loop through hostfile */

	} else {
	    err("ERROR ! Missing Host-File !\r\n");
	}

	fclose(host_file);

	if(status > 0) {  // this is why status must be set back to zero
	    sprintf(buffer, "ERROR ! No Host-Connection ! Tried %d hosts.\r", status);
	    err(buffer);
	}

	/*
	 * here the server connection is established !
	 */
	strcpy(buffer,revision);
	send(STDOUT_FILENO, buffer, strlen(buffer), 0);

	time(&tstart);
	log_logon(usercall,tstart,hostname);


	/*
	 * Read configfile, start with nodecall followed by ctext
	 */

	if((config_file = fopen(configfile,"r")) != NULL) {

	    if(fgets(nodecall, 20, config_file) != NULL) {

		if (strlen(nodecall) > 10) {
		    err("ERROR ! Invalid Node-Callsign !\r");
		} else {
		    cp = nodecall;
		    while(*cp != '\0') {
			*cp = toupper(*cp);
			if ((*cp == '\n') || (*cp == '\r')){
			    *cp = '\0';
			    break;
			}
			cp++;
		    }
		    if (verbose) {
			sprintf(buffer, "*** Nodecall: >%s<\r",nodecall);
			send(STDOUT_FILENO, buffer, strlen(buffer), 0);
		    }
		}
	    } else {
		err("ERROR IN CONFIG FILE - Missing Node-Callsign !\r");
	    }


	    while(fgets(buffer, 80, config_file) != NULL) {
		cp = buffer;
		while(*cp != '\0') {
		    if(*cp == '\n') {
			*cp = '\r';
		    }
		    cp++;
		}
		send(STDOUT_FILENO, buffer, strlen(buffer), 0);
	    }
	    fclose(config_file);

	} else {
	    err("ERROR ! Missing Config-File !\r\n");
	}

	if (verbose) {
		strcpy(buffer, "*** Connected\r");
		send(STDOUT_FILENO, buffer, strlen(buffer), 0);
	}

	userstat = select_loop(s, usercall, passwd, nodecall);

	if (verbose) {
		strcpy(buffer, "\r*** Disconnected\r");
		send(STDOUT_FILENO, buffer, strlen(buffer), 0);
	}
	time(&tstop);

	if(userstat.errcode >= 0) {
	    sprintf(buffer, "; Sent: %li Bytes / Received: %li Bytes, %.1fs, %.1f Bytes/s (%i)\r",userstat.bytes_sent, userstat.bytes_recv,difftime(tstop,tstart),(userstat.bytes_sent+userstat.bytes_recv)/difftime(tstop,tstart),userstat.errcode);
	    send(STDOUT_FILENO, buffer, strlen(buffer), 0);

	    sprintf(buffer, "; %s de %s SK\r",usercall,nodecall);
	    send(STDOUT_FILENO, buffer, strlen(buffer), 0);
	}

	log_logout(usercall,userstat,tstart,tstop);

	return 0;
}
